﻿using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;

using CancellationToken = System.Threading.CancellationToken;

namespace HeuristicLab.Algorithms.DataAnalysis.MoeaD {
  [Item("MOEA/D Algorithm", "Multi-objective Evolutionary Algorithm by Decomposition")]
  [StorableClass]
  [Creatable(CreatableAttribute.Categories.PopulationBasedAlgorithms, Priority = 125)]
  public class MOEADAlgorithm : MOEADAlgorithmBase {
    public MOEADAlgorithm() { }

    protected MOEADAlgorithm(MOEADAlgorithm original, Cloner cloner) : base(original, cloner) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new MOEADAlgorithm(this, cloner);
    }

    [StorableConstructor]
    protected MOEADAlgorithm(bool deserializing) : base(deserializing) { }

    protected override void Run(CancellationToken cancellationToken) {
      if (SetSeedRandomly) Seed = RandomSeedGenerator.GetSeed();
      var rand = RandomParameter.Value;
      rand.Reset(Seed);

      var functionTypeString = FunctionTypeParameter.Value.Value;
      switch (functionTypeString) {
        case "Chebyshev":
          functionType = FunctionType.TCHE;
          break;
        case "PBI":
          functionType = FunctionType.PBI;
          break;
        case "Weighted Sum":
          functionType = FunctionType.AGG;
          break;
      }

      var populationSize = PopulationSize.Value;
      population = new MOEADSolution[populationSize];

      bool[] maximization = ((BoolArray)Problem.MaximizationParameter.ActualValue).CloneAsArray();
      var dimensions = maximization.Length;

      IdealPoint = new double[dimensions];
      NadirPoint = new double[dimensions];

      var globalScope = new Scope("Global Scope");
      var executionContext = new ExecutionContext(null, this, globalScope);

      // set the execution context for parameters to allow lookup
      foreach (var parameter in Problem.Parameters.OfType<IValueParameter>()) {
        // we need all of these in order for the wiring of the operators to work
        globalScope.Variables.Add(new Variable(parameter.Name, parameter.Value));
      }
      globalScope.Variables.Add(new Variable("Results", Results)); // make results available as a parameter for analyzers etc.

      InitializePopulation(executionContext, cancellationToken, rand, maximization);
      IdealPoint.UpdateIdeal(population);
      NadirPoint.UpdateNadir(population);

      InitializeUniformWeights(rand, populationSize, dimensions);
      InitializeNeighbourHood(lambda, populationSize, NeighbourSize);

      var maximumEvaluatedSolutions = MaximumEvaluatedSolutions.Value;

      var crossover = Crossover;
      var crossoverProbability = CrossoverProbability.Value;
      var mutator = Mutator;
      var mutationProbability = MutationProbability.Value;
      var evaluator = Problem.Evaluator;
      var analyzer = Analyzer;
      var interpreter = Problem.SymbolicExpressionTreeInterpreter;
      var neighbourhoodSelectionProbability = NeighbourhoodSelectionProbability;

      previousBest = null;
      var evaluatedSolutions = populationSize;

      while (evaluatedSolutions < maximumEvaluatedSolutions) {
        foreach (var subProblemId in Enumerable.Range(0, populationSize).Shuffle(rand)) {
          var neighbourType = ChooseNeighborType(rand, neighbourhoodSelectionProbability);
          var mates = MatingSelection(rand, subProblemId, 2, neighbourType); // select parents
          var s1 = (IScope)population[mates[0]].Individual.Clone();
          var s2 = (IScope)population[mates[1]].Individual.Clone();
          s1.Parent = s2.Parent = globalScope;

          IScope childScope = null;
          // crossover
          if (rand.NextDouble() < crossoverProbability) {
            childScope = new Scope($"{mates[0]}+{mates[1]}") { Parent = executionContext.Scope };
            childScope.SubScopes.Add(s1);
            childScope.SubScopes.Add(s2);
            var op = executionContext.CreateChildOperation(crossover, childScope);
            ExecuteOperation(executionContext, cancellationToken, op);
            childScope.SubScopes.Clear();
          }

          // mutation
          if (rand.NextDouble() < mutationProbability) {
            childScope = childScope ?? s1;
            var op = executionContext.CreateChildOperation(mutator, childScope);
            ExecuteOperation(executionContext, cancellationToken, op);
          }

          // evaluation
          if (childScope != null) {
            var op = executionContext.CreateChildOperation(evaluator, childScope);
            ExecuteOperation(executionContext, cancellationToken, op);
            var qualities = (DoubleArray)childScope.Variables["Qualities"].Value;
            var childSolution = new MOEADSolution(childScope, dimensions, 0);
            // set child qualities
            for (int j = 0; j < dimensions; ++j) {
              childSolution.Qualities[j] = maximization[j] ? 1 - qualities[j] : qualities[j];
            }
            IdealPoint.UpdateIdeal(childSolution.Qualities);
            NadirPoint.UpdateNadir(childSolution.Qualities);
            // update neighbourhood will insert the child into the population
            UpdateNeighbourHood(rand, childSolution, subProblemId, neighbourType);

            ++evaluatedSolutions;
          } else {
            // no crossover or mutation were applied, a child was not produced, do nothing
          }

          if (evaluatedSolutions >= maximumEvaluatedSolutions) {
            break;
          }
        }
        // run analyzer
        var analyze = executionContext.CreateChildOperation(analyzer, globalScope);
        ExecuteOperation(executionContext, cancellationToken, analyze);

        UpdateBestSolution();
        UpdateParetoFronts();
        Results.AddOrUpdateResult("Evaluated Solutions", new IntValue(evaluatedSolutions));

        globalScope.SubScopes.Replace(population.Select(x => (IScope)x.Individual));
      }
    }

    private void InitializePopulation(ExecutionContext executionContext, CancellationToken cancellationToken, IRandom random, bool[] maximization) {
      var creator = Problem.SolutionCreator;
      var evaluator = Problem.Evaluator;

      var dimensions = maximization.Length;
      var populationSize = PopulationSize.Value;

      var parentScope = executionContext.Scope;
      // first, create all individuals
      for (int i = 0; i < populationSize; ++i) {
        var childScope = new Scope(i.ToString()) { Parent = parentScope };
        ExecuteOperation(executionContext, cancellationToken, executionContext.CreateChildOperation(creator, childScope));
        parentScope.SubScopes.Add(childScope);
      }

      // then, evaluate them and update qualities
      for (int i = 0; i < populationSize; ++i) {
        var childScope = parentScope.SubScopes[i];
        ExecuteOperation(executionContext, cancellationToken, executionContext.CreateChildOperation(evaluator, childScope));

        var qualities = (DoubleArray)childScope.Variables["Qualities"].Value;
        var solution = new MOEADSolution(childScope, dimensions, 0);
        for (int j = 0; j < dimensions; ++j) {
          solution.Qualities[j] = maximization[j] ? 1 - qualities[j] : qualities[j];
        }
        population[i] = solution;
      }
    }
  }
}
